# Define your item pipelines here
#
# Don't forget to add your pipeline to the ITEM_PIPELINES setting
# See: https://docs.scrapy.org/en/latest/topics/item-pipeline.html


# useful for handling different item types with a single interface
from datetime import datetime, timedelta
from itemadapter import ItemAdapter
from scrapy.exceptions import DropItem
from scrapy import Request
import logging
import json
import time
import os

import pymongo

from elasticsearch import Elasticsearch
from elasticsearch import helpers
from elasticsearch.client import IndicesClient
from elasticsearch.exceptions import RequestError
from elasticsearch.exceptions import ConnectionTimeout
from elasticsearch.exceptions import ConnectionError
from elasticsearch.exceptions import SSLError
from elasticsearch.exceptions import TransportError, ElasticsearchException, NotFoundError
from elasticsearch.helpers.errors import BulkIndexError

from dropshipping.utils.es_service import es_retry
from .logger import get_logger
from em_product.utils.models import *

def load_source_product_urls_by_sources(sources):
    config_path = os.path.join("~", ".em_celery", "config.ini")
    init_db(config_path)
    source_urls = {}
    batch_size = 4000
    offset = 0

    while True:
        products = (
            SpreeProductSource.select()
            .where(SpreeProductSource.source.in_(sources))
            .limit(batch_size)
            .offset(offset)
        )
        for product in products:
            source_urls[f"{product.source_product_id}"] = f"{product.source_product_url}"

        if len(products) < batch_size:
            break
        offset = offset + batch_size

    return source_urls

logger = get_logger(__name__, "app.log")


class MongoPipeline:
    BUFFER_SIZE = 1000

    def __init__(self, mongo_uri, mongo_db):
        pass

    def init_mongo_client(self, settings):
        pass

    @classmethod
    def from_crawler(cls, crawler):
        return cls(
            mongo_uri=crawler.settings.get("MONGO_URI"),
            mongo_db=crawler.settings.get("MONGO_DATABASE"),
        )


class ElasticSearchPipeline:
    ELASTICSEARCH_BUFFER_LENGTH = 500
    ELASTICSEARCH_INDEX = None
    FORCE = False
    ACTION = None

    def __init__(self, settings, id_key=None):
        self.id_key = id_key
        self.max_retry = 3
        self.init_es_client(settings)
        self.items_buffer = []

    def init_es_client(self, settings):
        es_timeout = settings.get("ELASTICSEARCH_TIMEOUT", 60)
        es_servers = settings.get("ELASTICSEARCH_SERVERS")
        es_servers = es_servers if isinstance(es_servers, list) else [es_servers]

        es_settings = dict()
        es_settings["hosts"] = es_servers
        es_settings["timeout"] = es_timeout

        if settings.get("ELASTICSEARCH_USERNAME") and settings.get("ELASTICSEARCH_PASSWORD"):
            es_settings["http_auth"] = (
                settings["ELASTICSEARCH_USERNAME"],
                settings["ELASTICSEARCH_PASSWORD"],
            )

        self.es = Elasticsearch(**es_settings)

    @classmethod
    def from_crawler(cls, crawler):
        return cls(crawler.settings)

    def update_item(self, item):
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        update_action = dict()
        if "_id" in item:
            update_action["_id"] = item.pop("_id")
        else:
            if self.id_key is not None and self.id_key in item:
                update_action["_id"] = item[self.id_key]
        update_action.update(
            {
                "_index": index_name,
                "_type": "_doc",
                "_op_type": "update",
                "doc_as_upsert": True,
                "doc": dict(item),
            }
        )

        self.items_buffer.append(update_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def save_item(self, item, action="create"):
        item.pop("_id", None)
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        index_action = {
            "_index": index_name,
            "_type": "_doc",
            "_op_type": action,
            "_source": dict(item),
        }

        if self.id_key is not None and self.id_key in item:
            index_action["_id"] = item[self.id_key]

        self.items_buffer.append(index_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def create_item(self, item):
        item.pop("_id", None)
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        index_action = {
            "_index": index_name,
            "_type": "_doc",
            "_op_type": "create",
            "_source": dict(item),
        }

        if self.id_key is not None and self.id_key in item:
            index_action["_id"] = item[self.id_key]

        self.items_buffer.append(index_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def index_item(self, item):
        item.pop("_id", None)
        index_name = item.pop("indice_name", self.ELASTICSEARCH_INDEX)
        index_action = {
            "_index": index_name,
            "_type": "_doc",
            "_op_type": "index",
            "_source": dict(item),
        }

        if self.id_key is not None and self.id_key in item:
            index_action["_id"] = item[self.id_key]

        self.items_buffer.append(index_action)

        if len(self.items_buffer) >= self.ELASTICSEARCH_BUFFER_LENGTH:
            self.send_items()
            self.items_buffer = []

    def send_items(self):
        retry = self.max_retry
        while retry > 0:
            try:
                res = helpers.bulk(self.es, self.items_buffer)
                logging.info(f"[Items Saved]: {self.items_buffer}")
                break
            except (ConnectionTimeout, ConnectionError, SSLError, TransportError):
                time.sleep(retry)
                retry -= 1
                continue
            except RequestError as e:
                break
            except BulkIndexError as e:
                for error in e.errors:
                    for _, value in error.items():
                        try:
                            logging.exception(value["_id"])
                        except Exception as e:
                            pass
                break
            except Exception as e:
                raise e

    def process_item(self, item, spider):
        if isinstance(item, list):
            for each in item:
                self.process_item(each, spider)
        else:
            if self.FORCE:
                self.save_item(item, "index")
            elif self.ACTION:
                self.save_item(item, self.ACTION)
            else:
                self.save_item(item)
            logging.debug("Item has been sent to Elastic Search %s", self.ELASTICSEARCH_INDEX)

        return item

    def ensure_indice(self, indice_name):
        ic = IndicesClient(self.es)
        if ic.exists(indice_name):
            return True

        ic.create(indice_name, {"settings": {"index.mapping.total_fields.limit": 100000}})
        return ic.exists(indice_name)

    def close_spider(self, spider):
        if len(self.items_buffer):
            self.send_items()


class ESCategoryPipeline(ElasticSearchPipeline):
    CATEGORY_INDEX = None
    PRODUCT_URL_KEY_NAME = "url"

    def __init__(self, settings, id_key="id"):
        super(ESCategoryPipeline, self).__init__(settings, id_key)

    def load_category_urls(self):
        params = {
            "index": self.CATEGORY_INDEX,
            "doc_type": "_doc",
            "size": 500,
            "query": {"query": {"match_all": {}}},
        }
        wrapped_scan = es_retry(helpers.scan)
        try:
            for item in wrapped_scan(self.es, **params):
                if isinstance(item["_source"], dict):
                    record = item["_source"]
                else:
                    record = json.loads(item["_source"])
                logging.debug(record)

                yield record
        except Exception as e:
            logging.exception(e)

    def open_spider(self, spider):
        if not self.ELASTICSEARCH_INDEX:
            raise Exception("Please provide es index")

        self.ensure_indice(self.CATEGORY_INDEX)
        self.ensure_indice(self.ELASTICSEARCH_INDEX)
        for doc in self.load_category_urls():
            if not doc.get(self.PRODUCT_URL_KEY_NAME):
                continue
            spider.start_urls.append(doc[self.PRODUCT_URL_KEY_NAME])

    def process_item(self, item, spider):
        if isinstance(item, list):
            for each in item:
                self.process_item(each, spider)
        else:
            self.save_item(item, "index")
            logging.debug("Item has been sent to Elastic Search %s", self.ELASTICSEARCH_INDEX)

        return item


class ESProductUrlPipeline(ESCategoryPipeline):
    def __init__(self, settings, id_key="id"):
        super(ESProductUrlPipeline, self).__init__(settings, id_key)


class ESProductPipeline(ElasticSearchPipeline):
    ACTION = None
    PRODUCT_URL_INDEX = None
    PRODUCT_INDEX = None
    PRODUCT_URL_ID_NAME = "id"
    PRODUCT_URL_KEY_NAME = "url"
    FORCE = False
    TTL = 7

    def __init__(self, settings):
        super(ESProductPipeline, self).__init__(settings, "product_id")

    def is_expired(self, product):
        if "date" not in product and "time" not in product:
            return True

        try:
            date_obj = datetime.fromisoformat(product.get("time", product.get("date")))
            return date_obj < datetime.now() - timedelta(days=self.TTL)
        except ValueError as e:
            return True

    def search_products(self, product_ids, key="product_id"):
        query = {"bool": {"filter": [{"terms": {key: product_ids}}]}}
        params = {
            "index": self.PRODUCT_INDEX,
            "from_": 0,
            "size": 500,
            "body": {"query": query},
        }

        resp = None
        retry = self.max_retry
        while retry > 0:
            try:
                resp = self.es.search(**params)
                break
            except NotFoundError as e:
                break
            except RequestError as e:
                raise e
            except ConnectionTimeout:
                time.sleep(1)
            except (TransportError, ElasticsearchException) as e:
                resp = -1
                retry -= 1

                status_code = getattr(e, "status_code", None)
                if status_code == "N/A":
                    time.sleep(3)
            except Exception as e:
                resp = -1
                retry -= 1

        if resp is None:
            result = None
        elif resp == -1:
            result = False
        else:
            result = dict()
            for item in resp["hits"]["hits"]:
                result.setdefault(item["_source"][key], [])
                result[item["_source"][key]].append(item["_source"])

        return result

    def load_product_urls(self):
        params = {
            "index": self.PRODUCT_URL_INDEX,
            "doc_type": "_doc",
            "size": 500,
            "query": {"query": {"match_all": {}}},
        }
        wrapped_scan = es_retry(helpers.scan)
        try:
            for item in wrapped_scan(self.es, **params):
                if isinstance(item["_source"], dict):
                    record = item["_source"]
                else:
                    record = json.loads(item["_source"])
                logging.debug(record)

                yield record
        except Exception as e:
            logging.exception(e)

    def get_urls(self, product_urls):
        if not self.FORCE:
            result = self.search_products(list(product_urls.keys()))
            if result:
                for id, product in result.items():
                    if not self.is_expired(product):
                        product_urls.pop(id)

        if product_urls:
            return [url for _, url in product_urls.items()]

    def open_spider(self, spider):
        if not self.ELASTICSEARCH_INDEX:
            raise Exception("Please provide es index")

        self.ensure_indice(self.PRODUCT_URL_INDEX)
        self.ensure_indice(self.ELASTICSEARCH_INDEX)
        batch_size = 500
        product_urls = dict()

        for doc in self.load_product_urls():
            if len(product_urls) < batch_size:
                product_urls[str(doc[self.PRODUCT_URL_ID_NAME])] = doc[self.PRODUCT_URL_KEY_NAME]
                continue

            urls = self.get_urls(product_urls)
            if urls:
                spider.start_urls.extend(urls)
            product_urls = {}

        if product_urls:
            urls = self.get_urls(product_urls)
            if urls:
                spider.start_urls.extend(urls)


class ESSeedProductPipeline(ElasticSearchPipeline):
    PRODUCT_URL_ID_NAME = "id"
    PRODUCT_URL_KEY_NAME = "url"
    FORCE = True

    def __init__(self, settings):
        super(ESSeedProductPipeline, self).__init__(settings, "product_id")

    def open_spider(self, spider):
        self.ensure_indice(self.ELASTICSEARCH_INDEX)


class ESProductRecrawlBySourcesPipeline(ESProductPipeline):
    PRODUCT_URL_INDEX = None
    PRODUCT_INDEX = None
    PRODUCT_URL_ID_NAME = "id"
    PRODUCT_URL_KEY_NAME = "url"
    FORCE = False
    SOURCES = None
    TTL = 7

    def __init__(self, settings):
        super(ESProductRecrawlBySourcesPipeline, self).__init__(settings)


    def open_spider(self, spider):
        if not self.ELASTICSEARCH_INDEX:
            logger.exception("[Index Not Found] {self.ELASTICSEARCH_INDEX}")
            raise Exception("Please provide es index")

        if not self.PRODUCT_INDEX:
            logger.exception("[Index Not Found] {self.PRODUCT_INDEX}")
            raise Exception("Please provide product es index")

        if self.SOURCES is None:
            raise Exception("Please provide sources")

        self.ensure_indice(self.ELASTICSEARCH_INDEX)
        source_urls = load_source_product_urls_by_sources(self.SOURCES)
        batch_size = 500
        product_urls = {}
        for id, url in source_urls.items():
            if len(product_urls) < batch_size:
                product_urls[str(id)] = url
                continue

            urls = self.get_urls(product_urls)
            if urls:
                spider.start_urls.extend(urls)
                logger.info(f"[Urls Added] {urls}")
            product_urls = {}

        if product_urls:
            urls = self.get_urls(product_urls)
            if urls:
                spider.start_urls.extend(urls)
                logger.info(f"[Urls Added] {urls}")

    def process_item(self, item, spider):
        if isinstance(item, list):
            for each in item:
                self.process_item(each, spider)
        else:
            if self.FORCE:
                self.index_item(item)
            else:
                self.update_item(item)
            logging.debug("Item has been sent to Elastic Search %s", self.ELASTICSEARCH_INDEX)

        return item

class ESProductRecrawlPipeline(ESProductPipeline):
    PRODUCT_URL_INDEX = None
    PRODUCT_INDEX = None
    PRODUCT_URL_ID_NAME = "id"
    PRODUCT_URL_KEY_NAME = "url"
    FORCE = False

    def __init__(self, settings):
        super(ESProductRecrawlPipeline, self).__init__(settings)

    def load_products(self, query=None):
        params = {
            "index": self.PRODUCT_INDEX,
            "doc_type": "_doc",
            "size": 500,
            "query": {"query": {"match_all": {}}},
        }
        if query is not None:
            params.update({"query": query})

        wrapped_scan = es_retry(helpers.scan)
        try:
            for item in wrapped_scan(self.es, **params):
                if isinstance(item["_source"], dict):
                    record = item["_source"]
                else:
                    record = json.loads(item["_source"])
                logging.debug(record)

                if not self.FORCE and not self.is_expired(record):
                    continue

                yield record
        except Exception as e:
            logging.exception(e)

    def open_spider(self, spider):
        if not self.ELASTICSEARCH_INDEX:
            logger.exception("[Index Not Found] {self.ELASTICSEARCH_INDEX}")
            raise Exception("Please provide es index")

        if not self.PRODUCT_INDEX:
            logger.exception("[Index Not Found] {self.PRODUCT_INDEX}")
            raise Exception("Please provide product es index")

        query = {
            "query": {
                "bool": {
                    "must_not": [{"exists": {"field": "options"}}],
                    "must": [{"exists": {"field": "existence"}}],
                }
            }
        }
        for doc in self.load_products(query):
            if doc.get(self.PRODUCT_URL_KEY_NAME):
                spider.start_urls.append(doc[self.PRODUCT_URL_KEY_NAME])
            logger.debug("[Url Added] {doc[self.PRODUCT_URL_KEY_NAME]}")

    def process_item(self, item, spider):
        if isinstance(item, list):
            for each in item:
                self.process_item(each, spider)
        else:
            self.update_item(item)
            logging.debug("Item has been sent to Elastic Search %s", self.ELASTICSEARCH_INDEX)

        return item


class ESProductUpdatePipeline(ElasticSearchPipeline):
    PRODUCT_URL_INDEX = None
    PRODUCT_INDEX = None
    PRODUCT_URL_ID_NAME = "id"
    PRODUCT_URL_KEY_NAME = "url"
    FORCE = False

    def open_spider(self, spider):
        if not self.ELASTICSEARCH_INDEX:
            logger.exception("[Index Not Found] {self.ELASTICSEARCH_INDEX}")
            raise Exception("Please provide es index")

        if not self.PRODUCT_INDEX:
            logger.exception("[Index Not Found] {self.PRODUCT_INDEX}")
            raise Exception("Please provide product es index")

    def process_item(self, item, spider):
        if isinstance(item, list):
            for each in item:
                self.process_item(each, spider)
        else:
            self.update_item(item)
            logging.debug("Item has been sent to Elastic Search %s", self.ELASTICSEARCH_INDEX)

        return item


class ESTranslationPipeline(ElasticSearchPipeline):
    PRODUCT_URL_INDEX = None
    PRODUCT_KEY_NAME = "id"
    TRANSLATION_INDEX = None

    def __init__(self, settings):
        self.translations = {}
        super(ESTranslationPipeline, self).__init__(settings, "id")

    def load_translation(self):
        params = {
            "index": self.TRANSLATION_INDEX,
            "doc_type": "_doc",
            "size": 500,
            "query": {"query": {"match_all": {}}},
        }
        wrapped_scan = es_retry(helpers.scan)
        try:
            for item in wrapped_scan(self.es, **params):
                if isinstance(item["_source"], dict):
                    record = item["_source"]
                else:
                    record = json.loads(item["_source"])
                logging.debug(record)

                yield record
        except Exception as e:
            logging.exception(e)

    def open_spider(self, spider):
        self.ensure_indice(self.TRANSLATION_INDEX)
        if not self.ELASTICSEARCH_INDEX:
            logger.exception("[Index Not Found] {self.ELASTICSEARCH_INDEX}")
            raise Exception("Please provide es index")

        if not self.TRANSLATION_INDEX:
            logger.exception("[Index Not Found] {self.PRODUCT_INDEX}")
            raise Exception("Please provide product es index")

        for doc in self.load_translation():
            if doc.get(self.PRODUCT_KEY_NAME):
                self.translations[self.PRODUCT_KEY_NAME] = doc

    def process_item(self, item, spider):
        if isinstance(item, list):
            for each in item:
                self.process_item(each, spider)
        else:
            self.create_item(item)
            logging.debug("Item has been sent to Elastic Search %s", self.ELASTICSEARCH_INDEX)

        return item
