import asyncio
from datetime import datetime
from pymongo import MongoClient, UpdateOne
from concurrent.futures import ThreadPoolExecutor


class MongoProduct(object):
    def __init__(self, host, port, user, password):
        self.host = host
        self.port = port
        self.user = user
        self.password = password

        mongo_uri = f"mongodb://{user}:{password}@{host}:{port}/?authSource=admin"
        self.client = MongoClient(mongo_uri)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.client.close()

    def load_yesstyle_products(self, options=None):
        return self.load_content("yesstyle", "products", options)

    def save_naver_product(self, product):
        self.save_naver_products([product])

    def save_naver_products(self, products):
        self.save_products("naver", "products", products)

    def save_trendyol_product(self, product):
        self.save_products("trendyol", "products", [product])

    def save_trendyol_product_url(self, product_url):
        self.save_products("trendyol", "product_urls", [product_url])

    def save_trendyol_products(self, products):
        self.save_products("trendyol", "products", products)

    def delete_product(self, db_name, collection_name, product_id):
        self.client[db_name][collection_name].delete_one({"_id": product_id})

    def delete_products(self, db_name, collection_name, product_ids):
        try:
            result = self.client[db_name][collection_name].delete_many({
                "_id": {"$in": product_ids}
            })
            print(f"Deleted {result.deleted_count} products.")
        except Exception as e:
            print(f"Error occurred while deleting products: {e}")


    def search_content(self, db_name, collection_name, ids):
        results = self.client[db_name][collection_name].find({"_id": {"$in": ids}})
        return {result["_id"]: [result] for result in results}

    def load_content(self, db_name, collection_name, options=None):
        query = {}
        if options:
            query.update(options)
        return self.client[db_name][collection_name].find(query, batch_size=4000)

    def save_products(self, db_name, collection_name, products):
        bulk_updates = []

        for product in products:
            if "_id" not in product and "id" in product:
                product["_id"] = product["id"]

            if "date" not in product:
                product["date"] = datetime.now().replace(microsecond=0)

            bulk_updates.append(
                UpdateOne({"_id": product.pop("_id")}, {"$set": product}, upsert=True)
            )
        with ThreadPoolExecutor(max_workers=10) as executor:
            asyncio.get_running_loop().run_in_executor(
                executor, self.client[db_name][collection_name].bulk_write, bulk_updates
            )
